#define MODEL

in vec4 inPosition;
in vec4 inColor;
in vec2 inTexCoord;
in float inFogCoord;
in vec3 inNormal;

out vec4 varColor;
out vec2 varTexCoord;
#ifdef USE_FOG
out float varFog;
#endif  // USE_FOG
#ifdef USE_ENV
out vec2 varEnvCoord;
#endif  // USE_ENV

// constants
BeginBlock(Constants)
_uniform vec3 shadowColor;
_uniform vec3 fogColor;
_uniform vec2 fogParams;
EndBlock

// projection matrix
BeginBlock(Projection)
_uniform mat4 matViewProj;
_uniform vec3 cameraPos;
_uniform vec3 cameraUpVec;
EndBlock

// params
BeginBlock(Params)
_uniform vec4 modelUpVec;
_uniform vec4 effectParams;
_uniform vec3 mirrorParams;
_uniform float alphaRef;
_uniform vec3 tint;
_uniform vec3 envColor;
_uniform vec3 envMatX;
_uniform vec3 envMatY;
_uniform bool envGood;
EndBlock

#ifdef USE_LIGHTS
void ProcessLights(void);
#endif  // USE_LIGHTS
#ifdef USE_EFFECTS
void ProcessEffects(inout vec4 pos);
#endif  // USE_EFFECTS

void main(void)
{
  vec4 pos = inPosition;

  varColor = inColor;
  varTexCoord = inTexCoord;

  #ifdef USE_EFFECTS
  // process effects
  ProcessEffects(pos);
  #endif  // USE_EFFECTS

  // calc position
  gl_Position = pos * matViewProj;

  #ifdef USE_FOG
  // is it in fog?
  float fogEnd = fogParams.x;
  float fogMul = fogParams.y;
  #ifdef MIRROR_FOG
  float mirrorHeight = mirrorParams.x;
  float mirrorMul = mirrorParams.y;
  float mirrorAdd = mirrorParams.z;
  float vertFog = dot(pos, modelUpVec);
  vertFog = mirrorHeight - vertFog;
  vertFog *= mirrorMul;
  vertFog += mirrorAdd;
  #else
  float vertFog = inFogCoord;
  #endif  // MIRROR_FOG
  varFog = (fogEnd - gl_Position.w) * fogMul;
  varFog = clamp(varFog, 0.0, 1.0);
  varFog -= vertFog;
  #endif  // USE_FOG

  #ifdef USE_ENV
  // calc env tex coords
  vec3 vecz = envGood ? normalize(pos.xyz - cameraPos) : vec3(0.0);
  vec3 vecx = envGood ? cross(cameraUpVec, vecz) * 0.5 : envMatX;
  vec3 vecy = envGood ? cross(vecz, vecx) * 0.5 : envMatY;
  varEnvCoord.s = dot(inNormal, vecx) + 0.5;
  varEnvCoord.t = dot(inNormal, vecy) + 0.5;
  #endif  // USE_ENV

  #ifdef USE_LIGHTS
  // process lights
  ProcessLights();
  #endif  // USE_LIGHTS

  #ifdef USE_EFFECTS
  // add any tinting
  varColor.rgb += tint;
  #endif  // USE_EFFECTS

  // clamp the color
  varColor = clamp(varColor, 0.0, 1.0);
}
